/* 
 * inferCP.c --- written by yasuko matsubara 2012.6.7
 */


#include "trimine.h"
#include "gamma.h"
#include "nrutil.h"
#include "tool.h"

void init_model();
void free_model();
void _samplingz(); 
void inference(); 
void _updatealpha();
void _updatebeta();
void _updategamma();
void calcParams();

#define   LHCHECK 	0 //1 //0 

void init_model(TRIMINE *model){

  model->alpha = 50.0/model->k;
  model->beta  = 0.1;
  model->gamma = 0.1;

  model->Nk  = (int*)  ivector(0, model->k);
  model->Nu  = (int*)  ivector(0, model->u);

  model->Nku = (int**) imatrix(0, model->k, 0, model->u); 
  model->Nkv = (int**) imatrix(0, model->k, 0, model->v); 
  model->Nkn = (int**) imatrix(0, model->k, 0, model->n); 

  model->O   = (double**) dmatrix(0, model->k, 0, model->u); 
  model->A   = (double**) dmatrix(0, model->k, 0, model->v); 
  model->C   = (double**) dmatrix(0, model->k, 0, model->n); 

  int i,k,t,j;
  for(i=0;i<model->k;i++)
    model->Nk[i] = 0;
  for(i=0;i<model->u;i++)
    model->Nu[i] = 0;
  for(k=0;k<model->k;k++){
    for(i=0;i<model->u;i++)
      model->Nku[k][i] = 0;
    for(i=0;i<model->v;i++)
      model->Nkv[k][i] = 0;
    for(i=0;i<model->n;i++)
      model->Nkn[k][i] = 0;
  }//k

  // init assign box
  for(t=0;t<model->n;t++){
    DATA *data = &model->tensor[t];
    data->z = (int*)  ivector(0, data->NM);
    for(i=0;i<data->NM;i++) data->z[i]=-1;
  }//t

  //compute nd, nt
  for(t=0;t<model->n;t++)
    for(i=0;i<model->u;i++)
      for(j=0;j<model->tensor[t].Mx[i];j++)
        model->Nu[i]+=model->tensor[t].x[i][j];

}

void free_model(TRIMINE *model){

  free_ivector(model->Nk, 0, model->k);
  free_ivector(model->Nu, 0, model->u);
  free_imatrix(model->Nku, 0, model->k, 0, model->u);
  free_imatrix(model->Nkv, 0, model->k, 0, model->v);
  free_imatrix(model->Nkn, 0, model->k, 0, model->n);

  free_dmatrix(model->O, 0, model->k, 0, model->u);
  free_dmatrix(model->A, 0, model->k, 0, model->v);
  free_dmatrix(model->C, 0, model->k, 0, model->n);

} 

void _samplingz(TRIMINE *model, int t){
  double posts[MAXK]; 
  int cnt=0; 
  int i,j,jj,l;

  DATA *data = &model->tensor[t];

  for(i=0;i<model->u;i++){
    for(j=0;j<data->Mx[i];j++){
      for(jj=0;jj<data->x[i][j];jj++){

	if(data->z[cnt] != -1){
	  model->Nk[data->z[cnt]]--;
	  model->Nkv[data->z[cnt]][data->ind[i][j]]--;
	  model->Nku[data->z[cnt]][i]--;
	  model->Nkn[data->z[cnt]][t]--;
	}//if

        double Z=0.0;
        for(l=0;l<model->k;l++){
          double O  = (model->Nku[l][i]+model->alpha);
          double Od = (model->Nu[i] + model->alpha * model->k);
          double A  = (model->Nkv[l][data->ind[i][j]]+model->beta);
          double Ad = (model->Nk[l] + model->beta * model->v);
          double C  = (model->Nkn[l][t]+model->gamma);
          double Cd = (model->Nk[l] + model->gamma * model->n);
          posts[l] = (O/Od)*(A/Ad)*(C/Cd);
          Z += posts[l];
        }//l 

        double u = (double)random()/ RAND_MAX ;
        u *= Z;
        int nextz = 0;
        double currprob = posts[0];

        while(u > currprob){
          nextz++;
          currprob += posts[nextz];
        }

        data->z[cnt] = nextz ;
        model->Nk[nextz] ++ ;
        model->Nkv[nextz][data->ind[i][j]] ++ ;
        model->Nku[nextz][i] ++ ;
        model->Nkn[nextz][t] ++ ;
        cnt ++ ;

      }//k
    }//j
  }//i

}



#if(LHCHECK)
double _likelihood(TRIMINE *model){

  int i,j;
  double L = model->u*lgamma(model->alpha*model->k)-model->u*model->k*lgamma(model->alpha);
  L += model->k*lgamma(model->beta*model->v)-model->k*model->v*lgamma(model->beta);
  for(i=0;i<model->u;i++){
    L -= lgamma(model->Nu[i]+model->alpha*model->k);
    for(j=0;j<model->k;j++)
      L += lgamma(model->Nku[j][i]+model->alpha);
  }//i
  for(i=0;i<model->k;i++){
    L -= lgamma(model->Nk[i]+model->beta*model->v);
    for(j=0;j<model->v;j++)
      L += lgamma(model->Nkv[i][j]+model->beta);
  }//i
  return L ;
}
#endif


void inference(TRIMINE *model, int flg){

  int i,j;

#if(LHCHECK)
  double preL = _likelihood(model) ;
#endif

  fprintf(stderr, "iter:");
  int maxstep = STEP ;
  for(i=0;i<maxstep;i++){
    fprintf(stderr, "%d ", i);
    for(j=0;j<model->n;j++)
      _samplingz(model, j);

    if( flg==1 && (i%ESTFREQ)==0 ){
      _updatealpha(model);
      _updatebeta(model);
      _updategamma(model);
    }//endif

#if(LHCHECK)
    double L = _likelihood(model) ;
    double v = L - preL ;
    fprintf(stderr, "%d \t %e \t %f \t %f \t %f \t %f\n", i+1, L, v, model->alpha, model->beta, model->gamma);
    preL = L ;
#endif
  }//i
}

void _updatealpha(TRIMINE *model){
  int i,j;
  double nume = - model->u*model->k*digamma(model->alpha);
  double deno = - model->u*model->k*digamma(model->alpha*model->k);
  for(i=0;i<model->u;i++){
    deno += model->k*digamma(model->Nu[i]+model->alpha*model->k);
    for(j=0;j<model->k;j++)
      nume += digamma(model->Nku[j][i]+model->alpha);
  }//i

  model->alpha *= nume/deno;
  double eps = 1e-8;
  if(model->alpha<eps) 
    model->alpha=eps;
}

void _updatebeta(TRIMINE *model){
  int i,j;
  double nume = - model->k*model->v*digamma(model->beta);
  double deno = - model->k*model->v*digamma(model->beta*model->v);
  for(i=0;i<model->k;i++){
    deno += model->v*digamma(model->Nk[i]+model->beta*model->v);
    for(j=0;j<model->v;j++) 
      nume += digamma(model->Nkv[i][j]+model->beta);
  }//i
  model->beta *= nume/deno;
  double eps = 1e-8;
  if(model->beta<eps) 
    model->beta=eps;
}

void _updategamma(TRIMINE *model){
  int i,j;
  double nume = - model->k*model->n*digamma(model->gamma);
  double deno = - model->k*model->n*digamma(model->gamma*model->n);
  for(i=0;i<model->k;i++){
    deno += model->n*digamma(model->Nk[i]+model->gamma*model->n);
    for(j=0;j<model->n;j++) 
      nume += digamma(model->Nkn[i][j]+model->gamma);
  }//i
  model->gamma *= nume/deno;
  double eps = 1e-8;
  if(model->gamma<eps) 
    model->gamma=eps;
}

double logsumexp(double x, double y, int flg){
  if(flg) return y; // init mode
  if(x == y) return x + 0.69314718055;  // log(2)
  double vmin, vmax;
  if(x > y) { vmax = x; vmin = y; }
  else{ vmax = y; vmin = x; }
  if(vmax > vmin + 50) 
    return vmax;
  else 
    return vmax + log (exp (vmin - vmax) + 1.0);
}

//--------------------------------//
// WRITE
//--------------------------------//
void calcParams(TRIMINE *model){
int i,k;

  for (k=0;k<model->k;k++){
    for(i=0;i<model->u;i++)
      model->O[k][i] = (model->Nku[k][i]+model->alpha)/(model->Nu[i]+model->alpha*model->k);
    for(i=0;i<model->v;i++)
      model->A[k][i] = (model->Nkv[k][i]+model->beta)/(model->Nk[k]+model->beta*model->v);
    for(i=0;i<model->n;i++)
      model->C[k][i] = (model->Nkn[k][i]+model->gamma)/(model->Nk[k]+model->gamma*model->n);
  }//k

}

void writeParams(char* dir, TRIMINE *model){
  char fn[BUFSIZE];
  // O & A & C matrix
  sprintf(fn, "%sO", dir);
  writeMatrix(fn, model->O, model->k, model->u);
  sprintf(fn, "%sA", dir);
  writeMatrix(fn, model->A, model->k, model->v);
  sprintf(fn, "%sC", dir);
  writeMatrix(fn, model->C, model->k, model->n);

}


