/* 
 * rls.c --- written by yasuko matsubara 2012.6.7
 */



#include        <stdio.h>
#include        <string.h>
#include        <stdlib.h>
#include        <math.h>
#include	"nrutil.h"
#include 	"tool.h"
#include 	"matrix.h"

//void rls(); 

void _RLS();
void _Malloc();
void _Mfrees();

/// computation space
typedef struct {
  int    w;
  double epsilon;
  double **gainM;
  double **identM;
  double **ataM;
  double *atbM;
  double  *_tmpV;
  double **_tmpM;
  double **_tmpM2;
  double **auxM;
} LS;


void rls(double **x, int w, int n, double *coeff){

  LS mats;
  mats.w = w;

  /// initialize the Gain matrix
  mats.epsilon = 0.0001;
  _Malloc(&mats, mats.w);
  identity(mats.gainM, w, 1.0/mats.epsilon);
  identity(mats.identM, w, 1.0);
  zeros(mats.ataM, w, w);
  zerosV(mats.atbM, w);

  //-------------------------//
  _RLS(&mats, x, n, coeff);
  //-------------------------//

  _Mfrees(&mats, mats.w);

}

#if(1==0)
int main(int argc, char* argv[]){
  int i,j;
  if(argc != 3) 
	error("Usage: ./cmd $dat order", argv[0]);
  char* fn = argv[1];
  int    w = atoi(argv[2]);
  int    n = count_lines(fn); 
  double **x; //x[n][w+1]
  x = (double**)dmatrix(0, n, 0, w+1);
  /// read input data
  FILE *fp;
  float fbuf;
  if((fp=fopen(fn,"r"))==NULL)
    error("cannot open",fn);
  for(j=0;j<n;j++){
    for(i=0;i<w+1;i++){
      fscanf(fp, "%f", &fbuf);
      x[j][i] = (double)fbuf;
    }//i
  }//j
  fclose(fp);
  /// recursive least squares
  double *coeff = (double*)dvector(0,w);
  rls(x, w, n, coeff);
  printVector(stderr, coeff, w);
}
#endif



void _RLS(LS *ma, double **x, int n, double *result){

  int i,j;

  int w = ma->w;
  double epsilon = ma->epsilon;
  double **gainM = ma->gainM; 
  double **identM = ma->identM;
  double **ataM   = ma->ataM;
  double *atbM    = ma->atbM;
  double  *_tmpV  = ma->_tmpV;
  double **_tmpM  = ma->_tmpM;
  double **_tmpM2 = ma->_tmpM2;
  double **auxM   = ma->auxM;

  for(i=0;i<n;i++){
    double *indVars = x[i];
    double *tV = indVars;
    double depVar = x[i][w];

    /// recursive least squares 
    // factor
    multVM(tV, gainM, w, w, _tmpV);
    double factor = 1.0 + multVV(tV, _tmpV, w);
    // auxM
    multVM(tV, gainM, w, w, _tmpV);
    outerproduct(tV, _tmpV, w, w, _tmpM);
    multNM(1.0/factor, _tmpM, w, w, auxM);
    // gainM
    minusMM(identM, auxM, w, w, _tmpM);
    multMM(gainM, _tmpM, w, w, _tmpM2);
    copyMM(_tmpM2, w,w, gainM);
    // ataM
    outerproduct(tV, tV, w,w, _tmpM);
    plusMM(ataM, _tmpM, w, w, _tmpM2);
    copyMM(_tmpM2, w,w, ataM);
    // update the At * b vector
    multNV(depVar, tV, w, _tmpV);
    plusVV(atbM, _tmpV, w, atbM);
    multVM(atbM, gainM, w, w, result);
  }//i

}
void _Malloc(LS *mats, int w){
  mats->gainM  = (double**) dmatrix(0, w, 0, w);
  mats->identM = (double**) dmatrix(0, w, 0, w);
  mats->ataM   = (double**) dmatrix(0, w, 0, w);
  mats->atbM   = (double*)  dvector(0, w);
  mats->_tmpV  = (double*)  dvector(0,w);
  mats->_tmpM  = (double**) dmatrix(0,w, 0,w);
  mats->_tmpM2 = (double**) dmatrix(0,w, 0,w);
  mats->auxM   = (double**) dmatrix(0,w, 0,w);
}

void _Mfrees(LS *mats, int w){
  free_dmatrix(mats->gainM,  0, w, 0, w);
  free_dmatrix(mats->identM, 0, w, 0, w);
  free_dmatrix(mats->ataM,   0, w, 0, w);
  free_dvector(mats->atbM,   0, w);
  free_dvector(mats->_tmpV,  0, w);
  free_dmatrix(mats->_tmpM,  0, w, 0, w);
  free_dmatrix(mats->_tmpM2, 0, w, 0, w);
  free_dmatrix(mats->auxM,   0, w, 0, w);
}

